"""
#lookup.py

This file contains all the mappings between hickle/HDF5 metadata and python
types.
There are three dictionaries that are populated here:

1) types_dict
Mapping between python types and dataset creation functions, e.g.
    types_dict = {
        list: (create_listlike_dataset, 'list'),
        int: (create_python_dtype_dataset, 'int'),
        np.ndarray: (create_np_array_dataset, 'ndarray'),
        }

2) hkl_types_dict
Mapping between hickle metadata and dataset loading functions, e.g.
    hkl_types_dict = {
        'list': load_list_dataset,
        'tuple': load_tuple_dataset
        }

3) dict_key_types_dict
Mapping specifically for converting hickled dict data back into a dictionary
with the same key type. While python dictionary keys can be any hashable
object, in HDF5 a unicode/string is required for a dataset name.

Example:
    dict_key_types_dict = {
        'str': literal_eval,
        'float': float
        }

## Extending hickle to add support for other classes and types

The process to add new load/dump capabilities is as follows:

1) Create a file called load_[newstuff].py in loaders/
2) In the load_[newstuff].py file, define your create_dataset and load_dataset
   functions, along with the 'class_register' and 'exclude_register' lists.

"""


# %% IMPORTS
# Built-in imports
from ast import literal_eval
from importlib import import_module
from inspect import isclass
from itertools import starmap


# %% GLOBALS
# Define dict of all acceptable types
types_dict = {}

# Define dict of all acceptable hickle types
hkl_types_dict = {}

# Define list of types that should never be sorted
types_not_to_sort = []

# Empty list of loaded loader names
loaded_loaders = []

# Define dict containing validation functions for ndarray-like objects
ndarray_like_check_fns = {}

# Define conversion dict of all acceptable dict key types
dict_key_types_dict = {
    b'str': literal_eval,
    b'float': float,
    b'bool': bool,
    b'int': int,
    b'complex': complex,
    b'tuple': literal_eval,
    b'NoneType': literal_eval,
    }


# %% FUNCTION DEFINITIONS
def load_nothing(h_node):
    pass


#####################
# ND-ARRAY checking #
#####################

def check_is_ndarray_like(py_obj):
    # Obtain the MRO of this object
    mro_list = py_obj.__class__.mro()

    # Create a function map
    func_map = map(ndarray_like_check_fns.get, mro_list)

    # Loop over the entire func_map until something else than None is found
    for func_item in func_map:
        if func_item is not None:
            return(func_item(py_obj))
    # If that did not happen, then py_obj is not ndarray_like
    else:
        return(False)


#####################
# loading optional  #
#####################

# This function registers a class to be used by hickle
def register_class(myclass_type, hkl_str, dump_function, load_function,
                   ndarray_check_fn=None, to_sort=True):
    """ Register a new hickle class.

    Args:
        myclass_type type(class): type of class
        hkl_str (str): String to write to HDF5 file to describe class
        dump_function (function def): function to write data to HDF5
        load_function (function def): function to load data from HDF5
        ndarray_check_fn (function def): function to use to check if
        to_sort (bool): If the item is iterable, does it require sorting?

    """
    types_dict[myclass_type] = (dump_function, hkl_str)
    hkl_types_dict[hkl_str] = load_function
    if not to_sort:
        types_not_to_sort.append(hkl_str)
    if ndarray_check_fn is not None:
        ndarray_like_check_fns[myclass_type] = ndarray_check_fn


def register_class_exclude(hkl_str_to_ignore):
    """ Tell loading funciton to ignore any HDF5 dataset with attribute
    'type=XYZ'

    Args:
        hkl_str_to_ignore (str): attribute type=string to ignore and exclude
            from loading.
    """
    hkl_types_dict[hkl_str_to_ignore] = load_nothing


# This function checks if an additional loader is required for given py_obj
def load_loader(py_obj):
    """
    Checks if given `py_obj` requires an additional loader to be handled
    properly and loads it if so.

    """

    # Obtain the MRO of this object
    if isclass(py_obj):
        mro_list = py_obj.mro()
    else:
        mro_list = py_obj.__class__.mro()

    # Loop over the entire mro_list
    for mro_item in mro_list:
        # Check if mro_item can be found in types_dict and return if so
        if mro_item in types_dict:
            return

        # Obtain the package name of mro_item
        pkg_name = mro_item.__module__.split('.')[0]

        # Obtain the name of the associated loader
        loader_name = 'hickle.loaders.load_%s' % (pkg_name)

        # Check if this module is already loaded, and return if so
        if loader_name in loaded_loaders:
            return

        # Try to load a loader with this name
        try:
            loader = import_module(loader_name)
        # If any module is not found, catch error and check it
        except ImportError as error:
            # Check if the error was due to a package in loader not being found
            if 'hickle' not in error.args[0]:   # pragma: no cover
                # If so, reraise the error
                raise
        # If such a loader does exist, register classes and return
        else:
            list(starmap(register_class, loader.class_register))
            list(map(register_class_exclude, loader.exclude_register))
            loaded_loaders.append(loader_name)
            return



# encoding: utf-8
"""
# load_python.py

Handlers for dumping and loading built-in python types.
NB: As these are for built-in types, they are critical to the functioning of
hickle.

"""


# %% IMPORTS
# Built-in imports
import warnings

# Package imports
import dill as pickle
import numpy as np

# hickle imports
from hickle.helpers import get_type_and_data


# %% CLASS DEFINITIONS
class SerializedWarning(UserWarning):
    """ An object type was not understood

    The data will be serialized using pickle.
    """
    pass


# %% FUNCTION DEFINITIONS
def create_listlike_dataset(py_obj, h_group, name, **kwargs):
    """ Dumper for list, set, tuple

    Args:
        py_obj: python object to dump; should be list-like
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    obj = list(py_obj)

    # h5py does not handle Py3 'str' objects well. Need to catch this
    # Only need to check first element as this method
    # is only called if all elements have same dtype
    str_type = None
    if type(obj[0]) in (str, bytes):
        str_type = bytes(type(obj[0]).__name__, 'ascii')

    if type(obj[0]) is str:
        obj = [bytes(oo, 'utf8') for oo in obj]

    d = h_group.create_dataset(name, data=obj, **kwargs)

    # Need to add some metadata to aid in unpickling if it's a string type
    if str_type is not None:
        d.attrs["str_type"] = str_type
    return(d)


def create_scalar_dataset(py_obj, h_group, name, **kwargs):
    """ dumps a python dtype object to h5py file

    Args:
        py_obj: python object to dump; should be a scalar (int, float,
            bool, str, etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    # Make sure 'compression' is not in kwargs
    kwargs.pop('compression', None)

    # If py_obj is an integer and cannot be stored in 64-bits, convert to str
    if isinstance(py_obj, int) and (py_obj.bit_length() > 64):
        py_obj = bytes(str(py_obj), 'ascii')

    d = h_group.create_dataset(name, data=py_obj, **kwargs)
    return(d)


def create_none_dataset(py_obj, h_group, name, **kwargs):
    """ Dump None type to file

    Args:
        py_obj: python object to dump; must be None object
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """
    d = h_group.create_dataset(name, data=b'None', **kwargs)
    return(d)


def create_pickled_dataset(py_obj, h_group, name, reason=None, **kwargs):
    """ If no match is made, raise a warning

    Args:
        py_obj: python object to dump; default if item is not matched.
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """
    reason_str = " (Reason: %s)" % (reason) if reason is not None else ""
    pickled_obj = pickle.dumps(py_obj)
    d = h_group.create_dataset(name, data=np.array(pickled_obj), **kwargs)

    warnings.warn("%r type not understood, data has been serialized%s"
                  % (py_obj.__class__.__name__, reason_str), SerializedWarning)
    return(d)


def load_list_dataset(h_node):
    _, _, data = get_type_and_data(h_node)
    str_type = h_node.attrs.get('str_type', None)

    if str_type == b'str':
        return(np.array(data, copy=False, dtype=str).tolist())
    else:
        return(data.tolist())


def load_tuple_dataset(h_node):
    data = load_list_dataset(h_node)
    return tuple(data)


def load_set_dataset(h_node):
    data = load_list_dataset(h_node)
    return set(data)


def load_none_dataset(h_node):
    return None


def load_pickled_data(h_node):
    _, _, data = get_type_and_data(h_node)
    return pickle.loads(data)


def load_scalar_dataset(h_node):
    _, base_type, data = get_type_and_data(h_node)

    if(base_type == b'int'):
        data = int(data)

    return(data)


# %% REGISTERS
class_register = [
    [list, b"list", create_listlike_dataset, load_list_dataset],
    [tuple, b"tuple", create_listlike_dataset, load_tuple_dataset],
    [set, b"set", create_listlike_dataset, load_set_dataset],
    [bytes, b"bytes", create_scalar_dataset, load_scalar_dataset],
    [str, b"str", create_scalar_dataset, load_scalar_dataset],
    [int, b"int", create_scalar_dataset, load_scalar_dataset],
    [float, b"float", create_scalar_dataset, load_scalar_dataset],
    [complex, b"complex", create_scalar_dataset, load_scalar_dataset],
    [bool, b"bool", create_scalar_dataset, load_scalar_dataset],
    [type(None), b"None", create_none_dataset, load_none_dataset],
    [object, b"pickle", create_pickled_dataset, load_pickled_data]]

exclude_register = []



# %% IMPORTS
# Package imports
from astropy.coordinates import Angle, SkyCoord
from astropy.constants import Constant
from astropy.table import Table
from astropy.time import Time
from astropy.units import Quantity
import numpy as np

# hickle imports
from hickle.helpers import get_type_and_data


# %% FUNCTION DEFINITIONS
def create_astropy_quantity(py_obj, h_group, name, **kwargs):
    """ dumps an astropy quantity

    Args:
        py_obj: python object to dump; should be a python type (int, float,
            bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    d = h_group.create_dataset(name, data=py_obj.value, dtype='float64',
                               **kwargs)
    unit = bytes(str(py_obj.unit), 'ascii')
    d.attrs['unit'] = unit
    return(d)


def create_astropy_angle(py_obj, h_group, name, **kwargs):
    """ dumps an astropy quantity

    Args:
        py_obj: python object to dump; should be a python type (int, float,
            bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    d = h_group.create_dataset(name, data=py_obj.value, dtype='float64',
                               **kwargs)
    unit = str(py_obj.unit).encode('ascii')
    d.attrs['unit'] = unit
    return(d)


def create_astropy_skycoord(py_obj, h_group, name, **kwargs):
    """ dumps an astropy quantity

    Args:
        py_obj: python object to dump; should be a python type (int, float,
            bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    lat = py_obj.data.lat.value
    lon = py_obj.data.lon.value
    dd = np.stack((lon, lat), axis=-1)

    d = h_group.create_dataset(name, data=dd, dtype='float64', **kwargs)
    lon_unit = str(py_obj.data.lon.unit).encode('ascii')
    lat_unit = str(py_obj.data.lat.unit).encode('ascii')
    d.attrs['lon_unit'] = lon_unit
    d.attrs['lat_unit'] = lat_unit
    return(d)


def create_astropy_time(py_obj, h_group, name, **kwargs):
    """ dumps an astropy Time object

    Args:
        py_obj: python object to dump; should be a python type (int, float,
            bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    data = py_obj.value
    dtype = str(py_obj.value.dtype)

    # Need to catch string times
    if '<U' in dtype:
        dtype = dtype.replace('<U', '|S')
        print(dtype)
        data = []
        for item in py_obj.value:
            data.append(str(item).encode('ascii'))

    d = h_group.create_dataset(name, data=data, dtype=dtype, **kwargs)
    fmt = str(py_obj.format).encode('ascii')
    scale = str(py_obj.scale).encode('ascii')
    d.attrs['format'] = fmt
    d.attrs['scale'] = scale

    return(d)


def create_astropy_constant(py_obj, h_group, name, **kwargs):
    """ dumps an astropy constant

    Args:
        py_obj: python object to dump; should be a python type (int, float,
            bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    d = h_group.create_dataset(name, data=py_obj.value, dtype='float64',
                               **kwargs)
    d.attrs["unit"] = str(py_obj.unit)
    d.attrs["abbrev"] = str(py_obj.abbrev)
    d.attrs["name"] = str(py_obj.name)
    d.attrs["reference"] = str(py_obj.reference)
    d.attrs["uncertainty"] = py_obj.uncertainty

    if py_obj.system:
        d.attrs["system"] = py_obj.system
    return(d)


def create_astropy_table(py_obj, h_group, name, **kwargs):
    """ Dump an astropy Table

    Args:
        py_obj: python object to dump; should be a python type (int, float,
            bool etc)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """
    data = py_obj.as_array()
    d = h_group.create_dataset(name, data=data, dtype=data.dtype, **kwargs)

    colnames = [bytes(cn, 'ascii') for cn in py_obj.colnames]
    d.attrs['colnames'] = colnames
    for key, value in py_obj.meta.items():
        d.attrs[key] = value
    return(d)


def load_astropy_quantity_dataset(h_node):
    py_type, _, data = get_type_and_data(h_node)
    unit = h_node.attrs["unit"]
    q = py_type(data, unit, copy=False)
    return q


def load_astropy_time_dataset(h_node):
    py_type, _, data = get_type_and_data(h_node)
    fmt = h_node.attrs["format"].decode('ascii')
    scale = h_node.attrs["scale"].decode('ascii')
    q = py_type(data, format=fmt, scale=scale)
    return q


def load_astropy_angle_dataset(h_node):
    py_type, _, data = get_type_and_data(h_node)
    unit = h_node.attrs["unit"]
    q = py_type(data, unit)
    return q


def load_astropy_skycoord_dataset(h_node):
    py_type, _, data = get_type_and_data(h_node)
    lon_unit = h_node.attrs["lon_unit"]
    lat_unit = h_node.attrs["lat_unit"]
    q = py_type(data[..., 0], data[..., 1], unit=(lon_unit, lat_unit))
    return q


def load_astropy_constant_dataset(h_node):
    py_type, _, data = get_type_and_data(h_node)
    unit = h_node.attrs["unit"]
    abbrev = h_node.attrs["abbrev"]
    name = h_node.attrs["name"]
    ref = h_node.attrs["reference"]
    unc = h_node.attrs["uncertainty"]

    system = None
    if "system" in h_node.attrs.keys():
        system = h_node.attrs["system"]

    c = py_type(abbrev, name, data, unit, unc, ref, system)
    return c


def load_astropy_table(h_node):
    py_type, _, data = get_type_and_data(h_node)
    metadata = dict(h_node.attrs.items())
    metadata.pop('type')
    metadata.pop('base_type')
    metadata.pop('colnames')

    colnames = [cn.decode('ascii') for cn in h_node.attrs["colnames"]]

    t = py_type(data, names=colnames, meta=metadata)
    return t


def check_is_astropy_table(py_obj):
    return isinstance(py_obj, Table)


def check_is_astropy_quantity_array(py_obj):
    if(isinstance(py_obj, (Quantity, Time, Angle, SkyCoord)) and
       not py_obj.isscalar):
        return(True)
    else:
        return(False)


# %% REGISTERS
class_register = [
    [Quantity, b'astropy_quantity', create_astropy_quantity,
     load_astropy_quantity_dataset, check_is_astropy_quantity_array],
    [Time, b'astropy_time', create_astropy_time, load_astropy_time_dataset,
     check_is_astropy_quantity_array],
    [Angle, b'astropy_angle', create_astropy_angle, load_astropy_angle_dataset,
     check_is_astropy_quantity_array],
    [SkyCoord, b'astropy_skycoord', create_astropy_skycoord,
     load_astropy_skycoord_dataset, check_is_astropy_quantity_array],
    [Constant, b'astropy_constant', create_astropy_constant,
     load_astropy_constant_dataset],
    [Table, b'astropy_table',  create_astropy_table, load_astropy_table,
     check_is_astropy_table]]

exclude_register = []



#! /usr/bin/env python
# encoding: utf-8
"""
# test_hickle_helpers.py

Unit tests for hickle module -- helper functions.

"""


# %% IMPORTS
# Package imports
import numpy as np

# hickle imports
from hickle.helpers import (
    check_is_hashable, check_is_iterable, check_iterable_item_type)
from hickle.loaders.load_numpy import check_is_numpy_array


# %% FUNCTION DEFINITIONS
def test_check_is_iterable():
    assert check_is_iterable([1, 2, 3])
    assert not check_is_iterable(1)


def test_check_is_hashable():
    assert check_is_hashable(1)
    assert not check_is_hashable([1, 2, 3])


def test_check_iterable_item_type():
    assert check_iterable_item_type([1, 2, 3]) is int
    assert not check_iterable_item_type([int(1), float(1)])
    assert not check_iterable_item_type([])


def test_check_is_numpy_array():
    assert check_is_numpy_array(np.array([1, 2, 3]))
    assert check_is_numpy_array(np.ma.array([1, 2, 3]))
    assert not check_is_numpy_array([1, 2])


# %% MAIN SCRIPT
if __name__ == "__main__":
    test_check_is_hashable()
    test_check_is_iterable()
    test_check_is_numpy_array()
    test_check_iterable_item_type()



# %% IMPORTS
# Package imports
import numpy as np
from py.path import local
from scipy.sparse import csr_matrix, csc_matrix, bsr_matrix

# hickle imports
import hickle
from hickle.loaders.load_scipy import check_is_scipy_sparse_array

# Set the current working directory to the temporary directory
local.get_temproot().chdir()


# %% FUNCTION DEFINITIONS
def test_is_sparse():
    sm0 = csr_matrix((3, 4), dtype=np.int8)
    sm1 = csc_matrix((1, 2))

    assert check_is_scipy_sparse_array(sm0)
    assert check_is_scipy_sparse_array(sm1)


def test_sparse_matrix():
    row = np.array([0, 0, 1, 2, 2, 2])
    col = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6])
    sm1 = csr_matrix((data, (row, col)), shape=(3, 3))
    sm2 = csc_matrix((data, (row, col)), shape=(3, 3))

    indptr = np.array([0, 2, 3, 6])
    indices = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6]).repeat(4).reshape(6, 2, 2)
    sm3 = bsr_matrix((data, indices, indptr), shape=(6, 6))

    hickle.dump(sm1, 'test_sp.h5')
    sm1_h = hickle.load('test_sp.h5')
    hickle.dump(sm2, 'test_sp2.h5')
    sm2_h = hickle.load('test_sp2.h5')
    hickle.dump(sm3, 'test_sp3.h5')
    sm3_h = hickle.load('test_sp3.h5')

    assert isinstance(sm1_h, csr_matrix)
    assert isinstance(sm2_h, csc_matrix)
    assert isinstance(sm3_h, bsr_matrix)

    assert np.allclose(sm1_h.data, sm1.data)
    assert np.allclose(sm2_h.data, sm2.data)
    assert np.allclose(sm3_h.data, sm3.data)

    assert sm1_h. shape == sm1.shape
    assert sm2_h. shape == sm2.shape
    assert sm3_h. shape == sm3.shape


# %% MAIN SCRIPT
if __name__ == "__main__":
    test_sparse_matrix()
    test_is_sparse()



# encoding: utf-8
"""
# load_numpy.py

Utilities and dump / load handlers for handling numpy and scipy arrays

"""

# %% IMPORTS
# Package imports
import numpy as np
import dill as pickle

# hickle imports
from hickle.helpers import get_type_and_data
from hickle.hickle import _dump


# %% FUNCTION DEFINITIONS
def check_is_numpy_array(py_obj):
    """ Check if a python object is a numpy array (masked or regular)

    Args:
        py_obj: python object to check whether it is a numpy array

    Returns
        is_numpy (bool): Returns True if it is a numpy array, else False if it
            isn't
    """

    return(isinstance(py_obj, np.ndarray))


def create_np_scalar_dataset(py_obj, h_group, name, **kwargs):
    """ dumps an np dtype object to h5py file

    Args:
        py_obj: python object to dump; should be a numpy scalar, e.g.
            np.float16(1)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    d = h_group.create_dataset(name, data=py_obj, **kwargs)

    d.attrs["np_dtype"] = bytes(str(d.dtype), 'ascii')
    return(d)


def create_np_dtype(py_obj, h_group, name, **kwargs):
    """ dumps an np dtype object to h5py file

    Args:
        py_obj: python object to dump; should be a numpy dtype, e.g.
            np.float16
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """
    d = h_group.create_dataset(name, data=str(py_obj), **kwargs)
    return(d)


def create_np_array_dataset(py_obj, h_group, name, **kwargs):
    """ dumps an ndarray object to h5py file

    Args:
        py_obj: python object to dump; should be a numpy array or np.ma.array
            (masked)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    # Obtain dtype of py_obj
    dtype = str(py_obj.dtype)

    # Check if py_obj contains strings
    if '<U' in dtype:
        # If so, convert the array to one with bytes
        py_obj = np.asanyarray(py_obj, dtype=dtype.replace('<U', '|S'))

    if isinstance(py_obj, np.ma.core.MaskedArray):
        d = h_group.create_dataset(name, data=py_obj, **kwargs)
        m = h_group.create_dataset('%s_mask' % name, data=py_obj.mask,
                                   **kwargs)
        m.attrs['type'] = np.array(pickle.dumps(py_obj.mask.__class__))
        m.attrs['base_type'] = b'ndarray_masked_mask'
    # Check if py_obj contains an object not understood by NumPy
    elif 'object' in dtype:
        # If so, convert py_obj to list
        py_obj = py_obj.tolist()

        # Check if py_obj is a list
        if isinstance(py_obj, list):
            # If so, dump py_obj into the current group
            _dump(py_obj, h_group, name, **kwargs)
            d = h_group[name]
            d.attrs['type'] = np.array(pickle.dumps(np.array))
        else:
            # If not, create a new group and dump py_obj into that
            d = h_group.create_group(name)
            _dump(py_obj, d, **kwargs)
            d.attrs['type'] = np.array(pickle.dumps(lambda x: np.array(x[0])))
    else:
        d = h_group.create_dataset(name, data=py_obj, **kwargs)
    d.attrs['np_dtype'] = bytes(dtype, 'ascii')
    return(d)


#####################
# Lookup dictionary #
#####################

def load_np_dtype_dataset(h_node):
    _, _, data = get_type_and_data(h_node)
    data = np.dtype(data)
    return data


def load_np_scalar_dataset(h_node):
    _, _, data = get_type_and_data(h_node)
    return data


def load_ndarray_dataset(h_node):
    _, _, data = get_type_and_data(h_node)
    dtype = h_node.attrs['np_dtype']
    return np.array(data, copy=False, dtype=dtype)


def load_ndarray_masked_dataset(h_node):
    _, _, data = get_type_and_data(h_node)
    dtype = h_node.attrs['np_dtype']
    try:
        mask_path = h_node.name + "_mask"
        h_root = h_node.parent
        mask = h_root.get(mask_path)[:]
    except (ValueError, IndexError):
        mask = h_root.get(mask_path)
    data = np.ma.array(data, mask=mask, dtype=dtype)
    return data


# %% REGISTERS
class_register = [
    [np.dtype, b"np_dtype", create_np_dtype, load_np_dtype_dataset],
    [np.ndarray, b"ndarray", create_np_array_dataset, load_ndarray_dataset,
     check_is_numpy_array],
    [np.ma.core.MaskedArray, b"ndarray_masked_data", create_np_array_dataset,
     load_ndarray_masked_dataset, check_is_numpy_array],
    [np.number, b"np_scalar", create_np_scalar_dataset,
     load_np_scalar_dataset]]

exclude_register = [b"ndarray_masked_mask"]



# %% IMPORTS
# Package imports
from astropy.units import Quantity
from astropy.time import Time
from astropy.coordinates import Angle, SkyCoord
import astropy.constants as apc
from astropy.table import Table
import numpy as np
from py.path import local

# hickle imports
import hickle as hkl

# Set the current working directory to the temporary directory
local.get_temproot().chdir()


# %% FUNCTION DEFINITIONS
def test_astropy_quantity():
    for uu in ['m^3', 'm^3 / s', 'kg/pc']:
        a = Quantity(7, unit=uu)

        hkl.dump(a, "test_ap.h5")
        b = hkl.load("test_ap.h5")

        assert a == b
        assert a.unit == b.unit

        a *= a
        hkl.dump(a, "test_ap.h5")
        b = hkl.load("test_ap.h5")
        assert a == b
        assert a.unit == b.unit


def test_astropy_constant():
    hkl.dump(apc.G, "test_ap.h5")
    gg = hkl.load("test_ap.h5")
    assert gg == apc.G

    hkl.dump(apc.cgs.e, 'test_ap.h5')
    ee = hkl.load('test_ap.h5')
    assert ee == apc.cgs.e


def test_astropy_table():
    t = Table([[1, 2], [3, 4]], names=('a', 'b'), meta={'name': 'test_thing'})

    hkl.dump({'a': t}, "test_ap.h5")
    t2 = hkl.load("test_ap.h5")['a']

    print(t)
    print(t.meta)
    print(t2)
    print(t2.meta)

    print(t.dtype, t2.dtype)
    assert t.meta == t2.meta
    assert t.dtype == t2.dtype

    assert np.allclose(t['a'].astype('float32'), t2['a'].astype('float32'))
    assert np.allclose(t['b'].astype('float32'), t2['b'].astype('float32'))


def test_astropy_quantity_array():
    a = Quantity([1, 2, 3], unit='m')

    hkl.dump(a, "test_ap.h5")
    b = hkl.load("test_ap.h5")

    assert np.allclose(a.value, b.value)
    assert a.unit == b.unit


def test_astropy_time_array():
    times = ['1999-01-01T00:00:00.123456789', '2010-01-01T00:00:00']
    t1 = Time(times, format='isot', scale='utc')
    hkl.dump(t1, "test_ap2.h5")
    t2 = hkl.load("test_ap2.h5")

    print(t1)
    print(t2)
    assert t1.value.shape == t2.value.shape
    for ii in range(len(t1)):
        assert t1.value[ii] == t2.value[ii]
    assert t1.format == t2.format
    assert t1.scale == t2.scale

    times = [58264, 58265, 58266]
    t1 = Time(times, format='mjd', scale='utc')
    hkl.dump(t1, "test_ap2.h5")
    t2 = hkl.load("test_ap2.h5")

    print(t1)
    print(t2)
    assert t1.value.shape == t2.value.shape
    assert np.allclose(t1.value, t2.value)
    assert t1.format == t2.format
    assert t1.scale == t2.scale


def test_astropy_angle():
    for uu in ['radian', 'degree']:
        a = Angle(1.02, unit=uu)

        hkl.dump(a, "test_ap.h5")
        b = hkl.load("test_ap.h5")
        assert a == b
        assert a.unit == b.unit


def test_astropy_angle_array():
    a = Angle([1, 2, 3], unit='degree')

    hkl.dump(a, "test_ap.h5")
    b = hkl.load("test_ap.h5")

    assert np.allclose(a.value, b.value)
    assert a.unit == b.unit


def test_astropy_skycoord():
    ra = Angle('1d20m', unit='degree')
    dec = Angle('33d0m0s', unit='degree')
    radec = SkyCoord(ra, dec)
    hkl.dump(radec, "test_ap.h5")
    radec2 = hkl.load("test_ap.h5")
    assert radec.ra == radec2.ra
    assert radec.dec == radec2.dec

    ra = Angle('1d20m', unit='hourangle')
    dec = Angle('33d0m0s', unit='degree')
    radec = SkyCoord(ra, dec)
    hkl.dump(radec, "test_ap.h5")
    radec2 = hkl.load("test_ap.h5")
    assert radec.ra == radec2.ra
    assert radec.dec == radec2.dec


def test_astropy_skycoord_array():
    ra = Angle(['1d20m', '0d21m'], unit='degree')
    dec = Angle(['33d0m0s', '-33d01m'], unit='degree')
    radec = SkyCoord(ra, dec)
    hkl.dump(radec, "test_ap.h5")
    radec2 = hkl.load("test_ap.h5")
    assert np.allclose(radec.ra.value, radec2.ra.value)
    assert np.allclose(radec.dec.value, radec2.dec.value)
    assert radec.ra.shape == radec2.ra.shape
    assert radec.dec.shape == radec2.dec.shape

    ra = Angle([['1d20m', '0d21m'], ['1d20m', '0d21m']], unit='hourangle')
    dec = Angle([['33d0m0s', '33d01m'], ['33d0m0s', '33d01m']], unit='degree')
    radec = SkyCoord(ra, dec)
    hkl.dump(radec, "test_ap.h5")
    radec2 = hkl.load("test_ap.h5")
    assert np.allclose(radec.ra.value, radec2.ra.value)
    assert np.allclose(radec.dec.value, radec2.dec.value)
    assert radec.ra.shape == radec2.ra.shape
    assert radec.dec.shape == radec2.dec.shape


# %% MAIN SCRIPT
if __name__ == "__main__":
    test_astropy_quantity()
    test_astropy_constant()
    test_astropy_table()
    test_astropy_quantity_array()
    test_astropy_time_array()
    test_astropy_angle()
    test_astropy_angle_array()
    test_astropy_skycoord()
    test_astropy_skycoord_array()



#! /usr/bin/env python
# encoding: utf-8
"""
# test_hickle.py

Unit tests for hickle module.

"""


# %% IMPORTS
# Built-in imports
from collections import OrderedDict as odict
import os
from pprint import pprint

# Package imports
import h5py
import numpy as np
from py.path import local
import pytest

# hickle imports
from hickle import dump, helpers, hickle, load, loaders

# Set current working directory to the temporary directory
local.get_temproot().chdir()


# %% GLOBALS
NESTED_DICT = {
    "level1_1": {
        "level2_1": [1, 2, 3],
        "level2_2": [4, 5, 6]
    },
    "level1_2": {
        "level2_1": [1, 2, 3],
        "level2_2": [4, 5, 6]
    },
    "level1_3": {
        "level2_1": {
            "level3_1": [1, 2, 3],
            "level3_2": [4, 5, 6]
        },
        "level2_2": [4, 5, 6]
    }
}


# %% HELPER DEFINITIONS
# Define a test function that must be serialized and unpacked again
def func(a, b, c=0):
    return(a, b, c)


# Define a class that must always be pickled
class with_state(object):
    def __init__(self):
        self.a = 12
        self.b = {
            'love': np.ones([12, 7]),
            'hatred': np.zeros([4, 9])}

    def __getstate__(self):
        self.a *= 2
        return({
            'a': self.a,
            'b': self.b})

    def __setstate__(self, state):
        self.a = state['a']
        self.b = state['b']

    def __getitem__(self, index):
        if(index == 0):
            return(self.a)
        if(index < 2):
            return(self.b['hatred'])
        if(index > 2):
            raise ValueError("index unknown")
        return(self.b['love'])


# %% FUNCTION DEFINITIONS
def test_invalid_file():
    """ Test if trying to use a non-file object fails. """

    with pytest.raises(hickle.FileError):
        dump('test', ())


def test_state_obj():
    """ Dumping and loading a class object with pickle states

    https://github.com/telegraphic/hickle/issues/125"""
    filename, mode = 'test.h5', 'w'
    obj = with_state()
    with pytest.warns(loaders.load_builtins.SerializedWarning):
        dump(obj, filename, mode)
    obj_hkl = load(filename)
    assert type(obj) == type(obj_hkl)
    assert np.allclose(obj[1], obj_hkl[1])


def test_local_func():
    """ Dumping and loading a local function

    https://github.com/telegraphic/hickle/issues/119"""
    filename, mode = 'test.h5', 'w'
    with pytest.warns(loaders.load_builtins.SerializedWarning):
        dump(func, filename, mode)
    func_hkl = load(filename)
    assert type(func) == type(func_hkl)
    assert func(1, 2) == func_hkl(1, 2)


def test_binary_file():
    """ Test if using a binary file works

    https://github.com/telegraphic/hickle/issues/123"""

    with open("test.hdf5", "w") as f:
        hickle.dump(None, f)

    with open("test.hdf5", "wb") as f:
        hickle.dump(None, f)


def test_non_empty_group():
    """ Test if attempting to dump to a group with data fails """

    hickle.dump(None, 'test.hdf5')
    with pytest.raises(ValueError):
        dump(None, 'test.hdf5', 'r+')


def test_string():
    """ Dumping and loading a string """
    filename, mode = 'test.h5', 'w'
    string_obj = "The quick brown fox jumps over the lazy dog"
    dump(string_obj, filename, mode)
    string_hkl = load(filename)
    assert isinstance(string_hkl, str)
    assert string_obj == string_hkl


def test_65bit_int():
    """ Dumping and loading an integer with arbitrary precision

    https://github.com/telegraphic/hickle/issues/113"""
    i = 2**65-1
    dump(i, 'test.hdf5')
    i_hkl = load('test.hdf5')
    assert i == i_hkl


def test_list():
    """ Dumping and loading a list """
    filename, mode = 'test_list.h5', 'w'
    list_obj = [1, 2, 3, 4, 5]
    dump(list_obj, filename, mode=mode)
    list_hkl = load(filename)
    try:
        assert isinstance(list_hkl, list)
        assert list_obj == list_hkl
        import h5py
        a = h5py.File(filename, 'r')
        a.close()

    except AssertionError:
        print("ERR:", list_obj, list_hkl)
        import h5py

        raise


def test_set():
    """ Dumping and loading a list """
    filename, mode = 'test_set.h5', 'w'
    list_obj = set([1, 0, 3, 4.5, 11.2])
    dump(list_obj, filename, mode)
    list_hkl = load(filename)
    try:
        assert isinstance(list_hkl, set)
        assert list_obj == list_hkl
    except AssertionError:
        print(type(list_obj))
        print(type(list_hkl))
        raise


def test_numpy():
    """ Dumping and loading numpy array """
    filename, mode = 'test.h5', 'w'
    dtypes = ['float32', 'float64', 'complex64', 'complex128']

    for dt in dtypes:
        array_obj = np.ones(8, dtype=dt)
        dump(array_obj, filename, mode)
        array_hkl = load(filename)
    try:
        assert array_hkl.dtype == array_obj.dtype
        assert np.all((array_hkl, array_obj))
    except AssertionError:
        print(array_hkl)
        print(array_obj)
        raise


def test_masked():
    """ Test masked numpy array """
    filename, mode = 'test.h5', 'w'
    a = np.ma.array([1, 2, 3, 4], dtype='float32', mask=[0, 1, 0, 0])

    dump(a, filename, mode)
    a_hkl = load(filename)

    try:
        assert a_hkl.dtype == a.dtype
        assert np.all((a_hkl, a))
    except AssertionError:
        print(a_hkl)
        print(a)
        raise


def test_object_numpy():
    """ Dumping and loading a NumPy array containing non-NumPy objects.

    https://github.com/telegraphic/hickle/issues/90"""

    arr = np.array([[NESTED_DICT], ('What is this?',), {1, 2, 3, 7, 1}])
    dump(arr, 'test.hdf5')
    arr_hkl = load('test.hdf5')
    assert np.all(arr == arr_hkl)

    arr2 = np.array(NESTED_DICT)
    dump(arr2, 'test.hdf5')
    arr_hkl2 = load('test.hdf5')
    assert np.all(arr2 == arr_hkl2)


def test_string_numpy():
    """ Dumping and loading NumPy arrays containing Python 3 strings. """

    arr = np.array(["1313e", "was", "maybe?", "here"])
    dump(arr, 'test.hdf5')
    arr_hkl = load('test.hdf5')
    assert np.all(arr == arr_hkl)


def test_list_object_numpy():
    """ Dumping and loading a list of NumPy arrays with objects.

    https://github.com/telegraphic/hickle/issues/90"""

    lst = [np.array(NESTED_DICT), np.array([('What is this?',),
                                            {1, 2, 3, 7, 1}])]
    dump(lst, 'test.hdf5')
    lst_hkl = load('test.hdf5')
    assert np.all(lst[0] == lst_hkl[0])
    assert np.all(lst[1] == lst_hkl[1])


def test_dict():
    """ Test dictionary dumping and loading """
    filename, mode = 'test.h5', 'w'

    dd = {
        'name': b'Danny',
        'age': 28,
        'height': 6.1,
        'dork': True,
        'nums': [1, 2, 3],
        'narr': np.array([1, 2, 3]),
    }

    dump(dd, filename, mode)
    dd_hkl = load(filename)

    for k in dd.keys():
        try:
            assert k in dd_hkl.keys()

            if isinstance(dd[k], np.ndarray):
                assert np.all((dd[k], dd_hkl[k]))
            else:
                pass
            assert isinstance(dd_hkl[k], dd[k].__class__)
        except AssertionError:
            print(k)
            print(dd_hkl[k])
            print(dd[k])
            print(type(dd_hkl[k]), type(dd[k]))
            raise


def test_odict():
    """ Test ordered dictionary dumping and loading

    https://github.com/telegraphic/hickle/issues/65"""
    filename, mode = 'test.hdf5', 'w'

    od = odict(((3, [3, 0.1]), (7, [5, 0.1]), (5, [3, 0.1])))
    dump(od, filename, mode)
    od_hkl = load(filename)

    assert od.keys() == od_hkl.keys()

    for od_item, od_hkl_item in zip(od.items(), od_hkl.items()):
        assert od_item == od_hkl_item


def test_empty_dict():
    """ Test empty dictionary dumping and loading

    https://github.com/telegraphic/hickle/issues/91"""
    filename, mode = 'test.h5', 'w'

    dump({}, filename, mode)
    assert load(filename) == {}


def test_compression():
    """ Test compression on datasets"""

    filename, mode = 'test.h5', 'w'
    dtypes = ['int32', 'float32', 'float64', 'complex64', 'complex128']

    comps = [None, 'gzip', 'lzf']

    for dt in dtypes:
        for cc in comps:
            array_obj = np.ones(32768, dtype=dt)
            dump(array_obj, filename, mode, compression=cc)
            print(cc, os.path.getsize(filename))
            array_hkl = load(filename)
    try:
        assert array_hkl.dtype == array_obj.dtype
        assert np.all((array_hkl, array_obj))
    except AssertionError:
        print(array_hkl)
        print(array_obj)
        raise


def test_dict_int_key():
    """ Test for dictionaries with integer keys """
    filename, mode = 'test.h5', 'w'

    dd = {
        0: "test",
        1: "test2"
    }

    dump(dd, filename, mode)
    load(filename)


def test_dict_nested():
    """ Test for dictionaries with integer keys """
    filename, mode = 'test.h5', 'w'

    dd = NESTED_DICT

    dump(dd, filename, mode)
    dd_hkl = load(filename)

    ll_hkl = dd_hkl["level1_3"]["level2_1"]["level3_1"]
    ll = dd["level1_3"]["level2_1"]["level3_1"]
    assert ll == ll_hkl


def test_masked_dict():
    """ Test dictionaries with masked arrays """

    filename, mode = 'test.h5', 'w'

    dd = {
        "data": np.ma.array([1, 2, 3], mask=[True, False, False]),
        "data2": np.array([1, 2, 3, 4, 5])
    }

    dump(dd, filename, mode)
    dd_hkl = load(filename)

    for k in dd.keys():
        try:
            assert k in dd_hkl.keys()
            if isinstance(dd[k], np.ndarray):
                assert np.all((dd[k], dd_hkl[k]))
            elif isinstance(dd[k], np.ma.MaskedArray):
                print(dd[k].data)
                print(dd_hkl[k].data)
                assert np.allclose(dd[k].data, dd_hkl[k].data)
                assert np.allclose(dd[k].mask, dd_hkl[k].mask)

            assert isinstance(dd_hkl[k], dd[k].__class__)

        except AssertionError:
            print(k)
            print(dd_hkl[k])
            print(dd[k])
            print(type(dd_hkl[k]), type(dd[k]))
            raise


def test_np_float():
    """ Test for singular np dtypes """
    filename, mode = 'np_float.h5', 'w'

    dtype_list = (np.float16, np.float32, np.float64,
                  np.complex64, np.complex128,
                  np.int8, np.int16, np.int32, np.int64,
                  np.uint8, np.uint16, np.uint32, np.uint64)

    for dt in dtype_list:

        dd = dt(1)
        dump(dd, filename, mode)
        dd_hkl = load(filename)
        assert dd == dd_hkl
        assert dd.dtype == dd_hkl.dtype

    dd = {}
    for dt in dtype_list:
        dd[str(dt)] = dt(1.0)
    dump(dd, filename, mode)
    dd_hkl = load(filename)

    print(dd)
    for dt in dtype_list:
        assert dd[str(dt)] == dd_hkl[str(dt)]


def test_comp_kwargs():
    """ Test compression with some kwargs for shuffle and chunking """

    filename, mode = 'test.h5', 'w'
    dtypes = ['int32', 'float32', 'float64', 'complex64', 'complex128']

    comps = [None, 'gzip', 'lzf']
    chunks = [(100, 100), (250, 250)]
    shuffles = [True, False]
    scaleoffsets = [0, 1, 2]

    for dt in dtypes:
        for cc in comps:
            for ch in chunks:
                for sh in shuffles:
                    for so in scaleoffsets:
                        kwargs = {
                            'compression': cc,
                            'dtype': dt,
                            'chunks': ch,
                            'shuffle': sh,
                            'scaleoffset': so
                        }
                        array_obj = NESTED_DICT
                        dump(array_obj, filename, mode, compression=cc)
                        print(kwargs, os.path.getsize(filename))
                        load(filename)


def test_list_numpy():
    """ Test converting a list of numpy arrays """

    filename, mode = 'test.h5', 'w'

    a = np.ones(1024)
    b = np.zeros(1000)
    c = [a, b]

    dump(c, filename, mode)
    dd_hkl = load(filename)

    print(dd_hkl)

    assert isinstance(dd_hkl, list)
    assert isinstance(dd_hkl[0], np.ndarray)


def test_tuple_numpy():
    """ Test converting a list of numpy arrays """

    filename, mode = 'test.h5', 'w'

    a = np.ones(1024)
    b = np.zeros(1000)
    c = (a, b, a)

    dump(c, filename, mode)
    dd_hkl = load(filename)

    print(dd_hkl)

    assert isinstance(dd_hkl, tuple)
    assert isinstance(dd_hkl[0], np.ndarray)


def test_numpy_dtype():
    """ Dumping and loading a NumPy dtype """

    dtype = np.dtype('float16')
    dump(dtype, 'test.hdf5')
    dtype_hkl = load('test.hdf5')
    assert dtype == dtype_hkl


def test_none():
    """ Test None type hickling """

    filename, mode = 'test.h5', 'w'

    a = None

    dump(a, filename, mode)
    dd_hkl = load(filename)
    print(a)
    print(dd_hkl)

    assert isinstance(dd_hkl, type(None))


def test_file_open_close():
    """ https://github.com/telegraphic/hickle/issues/20 """
    import h5py
    f = h5py.File('test.hdf', 'w')
    a = np.arange(5)

    dump(a, 'test.hkl')
    dump(a, 'test.hkl')

    dump(a, f, mode='w')
    f.close()
    try:
        dump(a, f, mode='w')
    except hickle.ClosedFileError:
        print("Tests: Closed file exception caught")


def test_hdf5_group():
    import h5py
    file = h5py.File('test.hdf5', 'w')
    group = file.create_group('test_group')
    a = np.arange(5)
    dump(a, group)
    file.close()

    a_hkl = load('test.hdf5', path='/test_group')
    assert np.allclose(a_hkl, a)

    file = h5py.File('test.hdf5', 'r+')
    group = file.create_group('test_group2')
    b = np.arange(8)

    dump(b, group, path='deeper/and_deeper')
    file.close()

    b_hkl = load('test.hdf5', path='/test_group2/deeper/and_deeper')
    assert np.allclose(b_hkl, b)

    file = h5py.File('test.hdf5', 'r')
    b_hkl2 = load(file['test_group2'], path='deeper/and_deeper')
    assert np.allclose(b_hkl2, b)
    file.close()


def test_list_order():
    """ https://github.com/telegraphic/hickle/issues/26 """
    d = [np.arange(n + 1) for n in range(20)]
    dump(d, 'test.h5')
    d_hkl = load('test.h5')

    try:
        for ii, xx in enumerate(d):
            assert d[ii].shape == d_hkl[ii].shape
        for ii, xx in enumerate(d):
            assert np.allclose(d[ii], d_hkl[ii])
    except AssertionError:
        print(d[ii], d_hkl[ii])
        raise


def test_embedded_array():
    """ See https://github.com/telegraphic/hickle/issues/24 """

    d_orig = [[np.array([10., 20.]), np.array([10, 20, 30])],
              [np.array([10, 2]), np.array([1.])]]
    dump(d_orig, 'test.h5')
    d_hkl = load('test.h5')

    for ii, xx in enumerate(d_orig):
        for jj, yy in enumerate(xx):
            assert np.allclose(d_orig[ii][jj], d_hkl[ii][jj])

    print(d_hkl)
    print(d_orig)


##############
# NEW TESTS  #
###############
def generate_nested():
    a = [1, 2, 3]
    b = [a, a, a]
    c = [a, b, 's']
    d = [a, b, c, c, a]
    e = [d, d, d, d, 1]
    f = {'a': a, 'b': b, 'e': e}
    g = {'f': f, 'a': e, 'd': d}
    h = {'h': g, 'g': f}
    z = [f, a, b, c, d, e, f, g, h, g, h]
    a = np.array([1, 2, 3, 4])
    b = set([1, 2, 3, 4, 5])
    c = (1, 2, 3, 4, 5)
    d = np.ma.array([1, 2, 3, 4, 5, 6, 7, 8])
    z = {'a': a, 'b': b, 'c': c, 'd': d, 'z': z}
    return z


def test_is_iterable():
    a = [1, 2, 3]
    b = 1

    assert helpers.check_is_iterable(a)
    assert not helpers.check_is_iterable(b)


def test_check_iterable_item_type():
    a = [1, 2, 3]
    b = [a, a, a]
    c = [a, b, 's']

    type_a = helpers.check_iterable_item_type(a)
    type_b = helpers.check_iterable_item_type(b)
    type_c = helpers.check_iterable_item_type(c)

    assert type_a is int
    assert type_b is list
    assert not type_c


def test_dump_nested():
    """ Dump a complicated nested object to HDF5
    """
    z = generate_nested()
    dump(z, 'test.hkl', mode='w')


def test_with_open_file():
    """
    Testing dumping and loading to an open file

    https://github.com/telegraphic/hickle/issues/92"""

    lst = [1]
    tpl = (1,)
    dct = {1: 1}
    arr = np.array([1])

    with h5py.File('test.hkl', 'w') as file:
        dump(lst, file, path='/lst')
        dump(tpl, file, path='/tpl')
        dump(dct, file, path='/dct')
        dump(arr, file, path='/arr')

    with h5py.File('test.hkl', 'r') as file:
        assert load(file, '/lst') == lst
        assert load(file, '/tpl') == tpl
        assert load(file, '/dct') == dct
        assert load(file, '/arr') == arr


def test_load():
    a = set([1, 2, 3, 4])
    b = set([5, 6, 7, 8])
    c = set([9, 10, 11, 12])
    z = (a, b, c)
    z = [z, z]
    z = (z, z, z, z, z)

    print("Original:")
    pprint(z)
    dump(z, 'test.hkl', mode='w')

    print("\nReconstructed:")
    z = load('test.hkl')
    pprint(z)


def test_sort_keys():
    keys = [b'data_0', b'data_1', b'data_2', b'data_3', b'data_10']
    keys_sorted = [b'data_0', b'data_1', b'data_2', b'data_3', b'data_10']

    print(keys)
    print(keys_sorted)
    assert helpers.sort_keys(keys) == keys_sorted


def test_ndarray():
    a = np.array([1, 2, 3])
    b = np.array([2, 3, 4])
    z = (a, b)

    print("Original:")
    pprint(z)
    dump(z, 'test.hkl', mode='w')

    print("\nReconstructed:")
    z = load('test.hkl')
    pprint(z)


def test_ndarray_masked():
    a = np.ma.array([1, 2, 3])
    b = np.ma.array([2, 3, 4], mask=[True, False, True])
    z = (a, b)

    print("Original:")
    pprint(z)
    dump(z, 'test.hkl', mode='w')

    print("\nReconstructed:")
    z = load('test.hkl')
    pprint(z)


def test_simple_dict():
    a = {'key1': 1, 'key2': 2}

    dump(a, 'test.hkl')
    z = load('test.hkl')

    pprint(a)
    pprint(z)


def test_complex_dict():
    a = {'akey': 1, 'akey2': 2}
    c = {'ckey': "hello", "ckey2": "hi there"}
    z = {'zkey1': a, 'zkey2': a, 'zkey3': c}

    print("Original:")
    pprint(z)
    dump(z, 'test.hkl', mode='w')

    print("\nReconstructed:")
    z = load('test.hkl')
    pprint(z)


def test_multi_hickle():
    """ Dumping to and loading from the same file several times

    https://github.com/telegraphic/hickle/issues/20"""

    a = {'a': 123, 'b': [1, 2, 4]}

    if os.path.exists("test.hkl"):
        os.remove("test.hkl")
    dump(a, "test.hkl", path="/test", mode="w")
    dump(a, "test.hkl", path="/test2", mode="r+")
    dump(a, "test.hkl", path="/test3", mode="r+")
    dump(a, "test.hkl", path="/test4", mode="r+")

    load("test.hkl", path="/test")
    load("test.hkl", path="/test2")
    load("test.hkl", path="/test3")
    load("test.hkl", path="/test4")


def test_complex():
    """ Test complex value dtype is handled correctly

    https://github.com/telegraphic/hickle/issues/29 """

    data = {"A": 1.5, "B": 1.5 + 1j, "C": np.linspace(0, 1, 4) + 2j}
    dump(data, "test.hkl")
    data2 = load("test.hkl")
    for key in data.keys():
        assert isinstance(data[key], data2[key].__class__)


def test_nonstring_keys():
    """ Test that keys are reconstructed back to their original datatypes
    https://github.com/telegraphic/hickle/issues/36
    """

    data = {
            u'test': 123,
            'def': [b'test'],
            'hik': np.array([1, 2, 3]),
            0: 0,
            True: ['test'],
            1.1: 'hey',
            1j: 'complex_hashable',
            (1, 2): 'boo',
            ('A', 17.4, 42): [1, 7, 'A'],
            (): '1313e was here',
            '0': 0,
            None: None
            }

    print(data)
    dump(data, "test.hkl")
    data2 = load("test.hkl")
    print(data2)

    for key in data.keys():
        assert key in data2.keys()

    print(data2)


def test_scalar_compression():
    """ Test bug where compression causes a crash on scalar datasets

    (Scalars are incompressible!)
    https://github.com/telegraphic/hickle/issues/37
    """
    data = {'a': 0, 'b': np.float(2), 'c': True}

    dump(data, "test.hkl", compression='gzip')
    data2 = load("test.hkl")

    print(data2)
    for key in data.keys():
        assert isinstance(data[key], data2[key].__class__)


def test_bytes():
    """ Dumping and loading a string. PYTHON3 ONLY """

    filename, mode = 'test.h5', 'w'
    string_obj = b"The quick brown fox jumps over the lazy dog"
    dump(string_obj, filename, mode)
    string_hkl = load(filename)
    print(type(string_obj))
    print(type(string_hkl))
    assert isinstance(string_hkl, bytes)
    assert string_obj == string_hkl


def test_np_scalar():
    """ Numpy scalar datatype

    https://github.com/telegraphic/hickle/issues/50
    """

    fid = 'test.h5py'
    r0 = {'test': np.float64(10.)}
    dump(r0, fid)
    r = load(fid)
    print(r)
    assert isinstance(r0['test'], r['test'].__class__)


def test_slash_dict_keys():
    """ Support for having slashes in dict keys

    https://github.com/telegraphic/hickle/issues/124"""
    dct = {'a/b': [1, '2'], 1.4: 3}

    dump(dct, 'test.hdf5', 'w')
    dct_hkl = load('test.hdf5')

    assert isinstance(dct_hkl, dict)
    for key, val in dct_hkl.items():
        assert val == dct.get(key)

    # Check that having backslashes in dict keys will serialize the dict
    dct2 = {'a\\b': [1, '2'], 1.4: 3}
    with pytest.warns(loaders.load_builtins.SerializedWarning):
        dump(dct2, 'test.hdf5')


# %% MAIN SCRIPT
if __name__ == '__main__':
    """ Some tests and examples """
    test_sort_keys()

    test_np_scalar()
    test_scalar_compression()
    test_complex()
    test_file_open_close()
    test_hdf5_group()
    test_none()
    test_masked_dict()
    test_list()
    test_set()
    test_numpy()
    test_dict()
    test_odict()
    test_empty_dict()
    test_compression()
    test_masked()
    test_dict_nested()
    test_comp_kwargs()
    test_list_numpy()
    test_tuple_numpy()
    test_list_order()
    test_embedded_array()
    test_np_float()
    test_string()
    test_nonstring_keys()
    test_bytes()

    # NEW TESTS
    test_is_iterable()
    test_check_iterable_item_type()
    test_dump_nested()
    test_with_open_file()
    test_load()
    test_sort_keys()
    test_ndarray()
    test_ndarray_masked()
    test_simple_dict()
    test_complex_dict()
    test_multi_hickle()
    test_dict_int_key()
    test_local_func()
    test_binary_file()
    test_state_obj()
    test_slash_dict_keys()
    test_invalid_file()
    test_non_empty_group()
    test_numpy_dtype()
    test_object_numpy()
    test_string_numpy()
    test_list_object_numpy()

    # Cleanup
    print("ALL TESTS PASSED!")



# %% IMPORTS
# Built-in imports
import glob
from os import path
import warnings

# Package imports
import h5py

# hickle imports
import hickle as hkl


# %% FUNCTION DEFINITIONS
def test_legacy_load():
    dirpath = path.dirname(__file__)
    filelist = sorted(glob.glob(path.join(dirpath, 'legacy_hkls/*.hkl')))

    # Make all warnings show
    warnings.simplefilter("always")

    for filename in filelist:
        try:
            print(filename)
            a = hkl.load(filename)
        except Exception:
            with h5py.File(filename) as a:
                print(a.attrs.items())
                print(a.items())
                for key, item in a.items():
                    print(item.attrs.items())
                raise


# %% MAIN SCRIPT
if __name__ == "__main__":
    test_legacy_load()



# %% IMPORTS
# Package imports
import dill as pickle
import numpy as np
import scipy
from scipy import sparse

# hickle imports
from hickle.helpers import get_type_and_data


# %% FUNCTION DEFINITIONS
def return_first(x):
    """ Return first element of a list """
    return x[0]


def check_is_scipy_sparse_array(py_obj):
    """ Check if a python object is a scipy sparse array

    Args:
        py_obj: python object to check whether it is a sparse array

    Returns
        is_numpy (bool): Returns True if it is a sparse array, else False if it
            isn't
    """
    t_csr = sparse.csr_matrix
    t_csc = sparse.csc_matrix
    t_bsr = sparse.bsr_matrix
    is_sparse = type(py_obj) in (t_csr, t_csc, t_bsr)

    return is_sparse


def create_sparse_dataset(py_obj, h_group, name, **kwargs):
    """ dumps an sparse array to h5py file

    Args:
        py_obj: python object to dump; should be a numpy array or np.ma.array
            (masked)
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """
    h_sparsegroup = h_group.create_group(name)
    data = h_sparsegroup.create_dataset('data', data=py_obj.data, **kwargs)
    indices = h_sparsegroup.create_dataset('indices', data=py_obj.indices,
                                           **kwargs)
    indptr = h_sparsegroup.create_dataset('indptr', data=py_obj.indptr,
                                          **kwargs)
    shape = h_sparsegroup.create_dataset('shape', data=py_obj.shape, **kwargs)

    if isinstance(py_obj, sparse.csr_matrix):
        type_str = 'csr'
    elif isinstance(py_obj, sparse.csc_matrix):
        type_str = 'csc'
    elif isinstance(py_obj, sparse.bsr_matrix):
        type_str = 'bsr'

    NoneType = type(None)
    h_sparsegroup.attrs['type'] = np.array(pickle.dumps(return_first))
    h_sparsegroup.attrs['base_type'] = ('%s_matrix' % type_str).encode('ascii')
    indices.attrs['type'] = np.array(pickle.dumps(NoneType))
    indices.attrs['base_type'] =\
        ("%s_matrix_indices" % type_str).encode('ascii')
    indptr.attrs['type'] = np.array(pickle.dumps(NoneType))
    indptr.attrs['base_type'] = ("%s_matrix_indptr" % type_str).encode('ascii')
    shape.attrs['type'] = np.array(pickle.dumps(NoneType))
    shape.attrs['base_type'] = ("%s_matrix_shape" % type_str).encode('ascii')

    return(data)


def load_sparse_matrix_data(h_node):
    _, base_type, data = get_type_and_data(h_node)
    h_root = h_node.parent
    indices = h_root.get('indices')[:]
    indptr = h_root.get('indptr')[:]
    shape = h_root.get('shape')[:]

    if base_type == b'csc_matrix':
        smat = sparse.csc_matrix((data, indices, indptr), dtype=data.dtype,
                                 shape=shape)
    elif base_type == b'csr_matrix':
        smat = sparse.csr_matrix((data, indices, indptr), dtype=data.dtype,
                                 shape=shape)
    elif base_type == b'bsr_matrix':
        smat = sparse.bsr_matrix((data, indices, indptr), dtype=data.dtype,
                                 shape=shape)
    return smat


# %% REGISTERS
class_register = [
    [scipy.sparse.csr_matrix, b'csr_matrix', create_sparse_dataset,
     load_sparse_matrix_data, check_is_scipy_sparse_array, False],
    [scipy.sparse.csc_matrix, b'csc_matrix', create_sparse_dataset,
     load_sparse_matrix_data, check_is_scipy_sparse_array, False],
    [scipy.sparse.bsr_matrix, b'bsr_matrix', create_sparse_dataset,
     load_sparse_matrix_data, check_is_scipy_sparse_array, False],
]

exclude_register = []

# Need to ignore things like csc_matrix_indices which are loaded automatically
for mat_type in ('csr', 'csc', 'bsr'):
    for attrib in ('indices', 'indptr', 'shape'):
        hkl_key = "%s_matrix_%s" % (mat_type, attrib)
        hkl_key = hkl_key.encode('ascii')
        exclude_register.append(hkl_key)



# encoding: utf-8
"""
# hickle.py

Created by Danny Price 2016-02-03.

Hickle is an HDF5 based clone of Pickle. Instead of serializing to a pickle
file, Hickle dumps to an HDF5 file. It is designed to be as similar to pickle
in usage as possible, providing a load() and dump() function.

## Notes

Hickle has two main advantages over Pickle:
1) LARGE PICKLE HANDLING. Unpickling a large pickle is slow, as the Unpickler
reads the entire pickle thing and loads it into memory. In comparison, HDF5
files are designed for large datasets. Things are only loaded when accessed.

2) CROSS PLATFORM SUPPORT. Attempting to unpickle a pickle pickled on Windows
on Linux and vice versa is likely to fail with errors like "Insecure string
pickle". HDF5 files will load fine, as long as both machines have
h5py installed.

"""


# %% IMPORTS
# Built-in imports
import io
from pathlib import Path
import sys
import warnings

# Package imports
import dill as pickle
import h5py as h5
import numpy as np

# hickle imports
from hickle import __version__
from hickle.helpers import (
    get_type, sort_keys, check_is_iterable, check_iterable_item_type)
from hickle.lookup import (
    types_dict, hkl_types_dict, types_not_to_sort, dict_key_types_dict,
    check_is_ndarray_like, load_loader)

# All declaration
__all__ = ['dump', 'load']


# %% CLASS DEFINITIONS
##################
# Error handling #
##################

class FileError(Exception):
    """ An exception raised if the file is fishy """
    pass


class ClosedFileError(Exception):
    """ An exception raised if the file is fishy """
    pass


class ToDoError(Exception):     # pragma: no cover
    """ An exception raised for non-implemented functionality"""
    def __str__(self):
        return "Error: this functionality hasn't been implemented yet."


# %% FUNCTION DEFINITIONS
def file_opener(f, path, mode='r'):
    """
    A file opener helper function with some error handling.
    This can open files through a file object, an h5py file, or just the
    filename.

    Parameters
    ----------
    f : file object, str or :obj:`~h5py.Group` object
        File to open for dumping or loading purposes.
        If str, `file_obj` provides the path of the HDF5-file that must be
        used.
        If :obj:`~h5py._hl.group.Group`, the group (or file) in an open
        HDF5-file that must be used.
    path : str
        Path within HDF5-file or group to dump to/load from.
    mode : str, optional
        Accepted values are 'r' (read only), 'w' (write; default) or 'a'
        (append).
        Ignored if file is a file object.

    """

    # Assume that we will have to close the file after dump or load
    close_flag = True

    # Make sure that the given path always starts with '/'
    if not path.startswith('/'):
        path = '/%s' % (path)

    # Were we handed a file object or just a file name string?
    if isinstance(f, (io.TextIOWrapper, io.BufferedWriter)):
        filename, mode = f.name, f.mode
        f.close()
        mode = mode.replace('b', '')
        h5f = h5.File(filename, mode)
    elif isinstance(f, (str, Path)):
        filename = f
        h5f = h5.File(filename, mode)
    elif isinstance(f, h5._hl.group.Group):
        try:
            filename = f.file.filename
        except ValueError:
            raise ClosedFileError("HDF5 file has been closed. Please pass "
                                  "either a filename string, a file object, or"
                                  "an open HDF5-file")
        path = ''.join([f.name, path])
        h5f = f.file

        if path.endswith('/'):
            path = path[:-1]

        # Since this file was already open, do not close the file afterward
        close_flag = False

    else:
        print(f.__class__)
        raise FileError("Cannot open file. Please pass either a filename "
                        "string, a file object, or a h5py.File")

    return(h5f, path, close_flag)


###########
# DUMPERS #
###########

# Get list of dumpable dtypes
dumpable_dtypes = [bool, complex, bytes, float, int, str]


def _dump(py_obj, h_group, call_id=None, **kwargs):
    """ Dump a python object to a group within an HDF5 file.

    This function is called recursively by the main dump() function.

    Args:
        py_obj: python object to dump.
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    # Check if we have a unloaded loader for the provided py_obj
    load_loader(py_obj)

    # Firstly, check if item is a numpy array. If so, just dump it.
    if check_is_ndarray_like(py_obj):
        create_hkl_dataset(py_obj, h_group, call_id, **kwargs)

    # Next, check if item is a dict
    elif isinstance(py_obj, dict):
        create_hkl_dataset(py_obj, h_group, call_id, **kwargs)

    # If not, check if item is iterable
    elif check_is_iterable(py_obj):
        item_type = check_iterable_item_type(py_obj)

        # item_type == False implies multiple types. Create a dataset
        if not item_type:
            h_subgroup = create_hkl_group(py_obj, h_group, call_id)
            for ii, py_subobj in enumerate(py_obj):
                _dump(py_subobj, h_subgroup, call_id=ii, **kwargs)

        # otherwise, subitems have same type. Check if subtype is an iterable
        # (e.g. list of lists), or not (e.g. list of ints, which should be
        # treated as a single dataset).
        else:
            if item_type in dumpable_dtypes:
                create_hkl_dataset(py_obj, h_group, call_id, **kwargs)
            else:
                h_subgroup = create_hkl_group(py_obj, h_group, call_id)
                for ii, py_subobj in enumerate(py_obj):
                    _dump(py_subobj, h_subgroup, call_id=ii, **kwargs)

    # item is not iterable, so create a dataset for it
    else:
        create_hkl_dataset(py_obj, h_group, call_id, **kwargs)


def dump(py_obj, file_obj, mode='w', path='/', **kwargs):
    """
    Write a hickled representation of `py_obj` to the provided `file_obj`.

    Parameters
    ----------
    py_obj : object
        Python object to hickle to HDF5.
    file_obj : file object, str or :obj:`~h5py.Group` object
        File in which to store the object.
        If str, `file_obj` provides the path of the HDF5-file that must be
        used.
        If :obj:`~h5py._hl.group.Group`, the group (or file) in an open
        HDF5-file that must be used.
    mode : str, optional
        Accepted values are 'r' (read only), 'w' (write; default) or 'a'
        (append).
        Ignored if file is a file object.
    path : str, optional
        Path within HDF5-file or group to save data to.
        Defaults to root ('/').
    kwargs : keyword arguments
        Additional keyword arguments that must be provided to the
        :meth:`~h5py._hl.group.Group.create_dataset` method.

    """

    # Make sure that file is not closed unless modified
    # This is to avoid trying to close a file that was never opened
    close_flag = False

    try:
        # Open the file
        h5f, path, close_flag = file_opener(file_obj, path, mode)

        # Log which version of python was used to generate the hickle file
        pv = sys.version_info
        py_ver = "%i.%i.%i" % (pv[0], pv[1], pv[2])

        # Try to create the root group
        try:
            h_root_group = h5f.create_group(path)

        # If that is not possible, check if it is empty
        except ValueError as error:
            # Raise error if this group is not empty
            if len(h5f[path]):
                raise error
            else:
                h_root_group = h5f.get(path)

        h_root_group.attrs["HICKLE_VERSION"] = __version__
        h_root_group.attrs["HICKLE_PYTHON_VERSION"] = py_ver

        _dump(py_obj, h_root_group, **kwargs)
    finally:
        # Close the file if requested.
        # Closing a file twice will not cause any problems
        if close_flag:
            h5f.close()


def create_dataset_lookup(py_obj):
    """ What type of object are we trying to hickle?  This is a python
    dictionary based equivalent of a case statement.  It returns the correct
    helper function for a given data type.

    Args:
        py_obj: python object to look-up what function to use to dump to disk

    Returns:
        match: function that should be used to dump data to a new dataset
        base_type: the base type of the data that will be dumped
    """

    # Obtain the MRO of this object
    mro_list = py_obj.__class__.mro()

    # Create a type_map
    type_map = map(types_dict.get, mro_list)

    # Loop over the entire type_map until something else than None is found
    for type_item in type_map:
        if type_item is not None:
            return(type_item)


def create_hkl_dataset(py_obj, h_group, call_id=None, **kwargs):
    """ Create a dataset within the hickle HDF5 file

    Args:
        py_obj: python object to dump.
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.

    """
    # lookup dataset creator type based on python object type
    create_dataset, base_type = create_dataset_lookup(py_obj)

    # Set the name of this dataset
    name = 'data%s' % ("_%i" % (call_id) if call_id is not None else '')

    # Try to create the dataset
    try:
        h_subgroup = create_dataset(py_obj, h_group, name, **kwargs)
    # If that fails, pickle the object instead
    except Exception as error:
        # Make sure builtins loader is loaded
        load_loader(object)

        # Obtain the proper dataset creator and base type
        create_dataset, base_type = types_dict[object]

        # Make sure that a group/dataset with name 'name' does not exist
        try:
            del h_group[name]
        except Exception:
            pass

        # Create the pickled dataset
        h_subgroup = create_dataset(py_obj, h_group, name, error, **kwargs)

    # Save base type of py_obj
    h_subgroup.attrs['base_type'] = base_type

    # Save a pickled version of the true type of py_obj if necessary
    if base_type != b'pickle' and 'type' not in h_subgroup.attrs:
        h_subgroup.attrs['type'] = np.array(pickle.dumps(py_obj.__class__))


def create_hkl_group(py_obj, h_group, call_id=None):
    """ Create a new group within the hickle file

    Args:
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.

    """

    # Set the name of this group
    if isinstance(call_id, str):
        name = call_id
    else:
        name = 'data%s' % ("_%i" % (call_id) if call_id is not None else '')

    h_subgroup = h_group.create_group(name)
    h_subgroup.attrs['type'] = np.array(pickle.dumps(py_obj.__class__))
    h_subgroup.attrs['base_type'] = create_dataset_lookup(py_obj)[1]
    return h_subgroup


def create_dict_dataset(py_obj, h_group, name, **kwargs):
    """ Creates a data group for each key in dictionary

    Notes:
        This is a very important function which uses the recursive _dump
        method to build up hierarchical data models stored in the HDF5 file.
        As this is critical to functioning, it is kept in the main hickle.py
        file instead of in the loaders/ directory.

    Args:
        py_obj: python object to dump; should be dictionary
        h_group (h5.File.group): group to dump data into.
        call_id (int): index to identify object's relative location in the
            iterable.
    """

    h_dictgroup = h_group.create_group(name)

    for idx, (key, py_subobj) in enumerate(py_obj.items()):
        # Obtain the raw string representation of this key
        subgroup_key = "%r" % (key)

        # Make sure that the '\\\\' is not in the key, or raise error if so
        if '\\\\' in subgroup_key:
            del h_group[name]
            raise ValueError("Dict item keys containing the '\\\\' string are "
                             "not supported!")

        # Replace any forward slashes with double backslashes
        subgroup_key = subgroup_key.replace('/', '\\\\')
        h_subgroup = h_dictgroup.create_group(subgroup_key)
        h_subgroup.attrs['base_type'] = b'dict_item'

        h_subgroup.attrs['key_base_type'] = bytes(type(key).__name__, 'ascii')
        h_subgroup.attrs['key_type'] = np.array(pickle.dumps(key.__class__))

        h_subgroup.attrs['key_idx'] = idx

        _dump(py_subobj, h_subgroup, call_id=None, **kwargs)
    return(h_dictgroup)


# Add create_dict_dataset to types_dict
types_dict[dict] = (create_dict_dataset, b"dict")


###########
# LOADERS #
###########

class PyContainer(list):
    """ A group-like object into which to load datasets.

    In order to build up a tree-like structure, we need to be able
    to load datasets into a container with an append() method.
    Python tuples and sets do not allow this. This class provides
    a list-like object that be converted into a list, tuple, set or dict.
    """
    def __init__(self):
        super(PyContainer, self).__init__()
        self.container_type = None
        self.container_base_type = None
        self.name = None
        self.key_type = None
        self.key_base_type = None

    def convert(self):
        """ Convert from PyContainer to python core data type.

        Returns: self, either as a list, tuple, set or dict
                 (or other type specified in lookup.py)
        """

        # If this container is a dict, convert its items properly
        if self.container_base_type == b"dict":
            # Create empty list of items
            items = [[]]*len(self)

            # Loop over all items in the container
            for item in self:
                # Obtain the name of this item
                key = item.name.split('/')[-1].replace('\\\\', '/')

                # Obtain the base type and index of this item's key
                key_base_type = item.key_base_type
                key_idx = item.key_idx

                # If this key has a type that must be converted, do so
                if key_base_type in dict_key_types_dict.keys():
                    to_type_fn = dict_key_types_dict[key_base_type]
                    key = to_type_fn(key)

                # Insert item at the correct index into the list
                items[key_idx] = [key, item[0]]

            # Initialize dict using its true type and return
            return(self.container_type(items))

        # In all other cases, return container
        else:
            # If container has a true type defined, convert to that first
            if self.container_type is not None:
                return(self.container_type(self))

            # If not, return the container itself
            else:
                return(self)


def no_match_load(key):     # pragma: no cover
    """ If no match is made when loading, need to raise an exception
    """
    raise RuntimeError("Cannot load %s data type" % key)


def load_dataset_lookup(key):
    """ What type of object are we trying to unpickle?  This is a python
    dictionary based equivalent of a case statement.  It returns the type
    a given 'type' keyword in the hickle file.

    Args:
        py_obj: python object to look-up what function to use to dump to disk

    Returns:
        match: function that should be used to dump data to a new dataset
    """

    match = hkl_types_dict.get(key, no_match_load)

    return match


def load(file_obj, path='/', safe=True):
    """
    Load the Python object stored in `file_obj` at `path` and return it.

    Parameters
    ----------
    file_obj : file object, str or :obj:`~h5py.Group` object
        File from which to load the object.
        If str, `file_obj` provides the path of the HDF5-file that must be
        used.
        If :obj:`~h5py._hl.group.Group`, the group (or file) in an open
        HDF5-file that must be used.
    path : str, optional
        Path within HDF5-file or group to load data from.
        Defaults to root ('/').
    safe : bool, optional
        Disable automatic depickling of arbitrary python objects.
        DO NOT set this to False unless the file is from a trusted source.
        (See https://docs.python.org/3/library/pickle.html for an explanation)

    Returns
    -------
    py_obj : object
        The unhickled Python object.

    """

    # Make sure that the file is not closed unless modified
    # This is to avoid trying to close a file that was never opened
    close_flag = False

    # Try to read the provided file_obj as a hickle file
    try:
        h5f, path, close_flag = file_opener(file_obj, path, 'r')
        h_root_group = h5f.get(path)

        # Define attributes h_root_group must have
        v3_attrs = ['CLASS', 'VERSION', 'PYTHON_VERSION']
        v4_attrs = ['HICKLE_VERSION', 'HICKLE_PYTHON_VERSION']

        # Check if the proper attributes for v3 loading are available
        if all(map(h_root_group.attrs.get, v3_attrs)):
            # Check if group attribute 'CLASS' has value 'hickle
            if(h_root_group.attrs['CLASS'] != b'hickle'):  # pragma: no cover
                # If not, raise error
                raise AttributeError("HDF5-file attribute 'CLASS' does not "
                                     "have value 'hickle'!")

            # Obtain version with which the file was made
            try:
                major_version = int(h_root_group.attrs['VERSION'][0])

            # If this cannot be done, then this is not a v3 file
            except Exception:  # pragma: no cover
                raise Exception("This file does not appear to be a hickle v3 "
                                "file.")

            # Else, if the major version is not 3, it is not a v3 file either
            else:
                if(major_version != 3):  # pragma: no cover
                    raise Exception("This file does not appear to be a hickle "
                                    "v3 file.")

            # Load file
            from hickle import legacy_v3
            warnings.warn("Input argument 'file_obj' appears to be a file made"
                          " with hickle v3. Using legacy load...")
            return(legacy_v3.load(file_obj, path, safe))

        # Else, check if the proper attributes for v4 loading are available
        elif all(map(h_root_group.attrs.get, v4_attrs)):
            # Load file
            py_container = PyContainer()
            py_container = _load(py_container, h_root_group['data'])
            return(py_container[0])

        # Else, raise error
        else:  # pragma: no cover
            raise FileError("HDF5-file does not have the proper attributes!")

    # If this fails, raise error and provide user with caught error message
    except Exception as error:  # pragma: no cover
        raise ValueError("Provided argument 'file_obj' does not appear to be a"
                         " valid hickle file! (%s)" % (error))
    finally:
        # Close the file if requested.
        # Closing a file twice will not cause any problems
        if close_flag:
            h5f.close()


def load_dataset(h_node):
    """ Load a dataset, converting into its correct python type

    Args:
        h_node (h5py dataset): h5py dataset object to read

    Returns:
        data: reconstructed python object from loaded data
    """
    py_type, base_type = get_type(h_node)
    load_loader(py_type)

    load_fn = load_dataset_lookup(base_type)
    data = load_fn(h_node)

    # If data is not py_type yet, convert to it (unless it is pickle)
    if base_type != b'pickle' and type(data) != py_type:
        data = py_type(data)
    return data


def _load(py_container, h_group):
    """ Load a hickle file

    Recursive funnction to load hdf5 data into a PyContainer()

    Args:
        py_container (PyContainer): Python container to load data into
        h_group (h5 group or dataset): h5py object, group or dataset, to spider
            and load all datasets.
    """

    # Either a file, group, or dataset
    if isinstance(h_group, h5._hl.group.Group):

        py_subcontainer = PyContainer()
        py_subcontainer.container_base_type = bytes(h_group.attrs['base_type'])

        py_subcontainer.name = h_group.name

        if py_subcontainer.container_base_type == b'dict_item':
            py_subcontainer.key_base_type = h_group.attrs['key_base_type']
            py_obj_type = pickle.loads(h_group.attrs['key_type'])
            py_subcontainer.key_type = py_obj_type
            py_subcontainer.key_idx = h_group.attrs['key_idx']
        else:
            py_obj_type = pickle.loads(h_group.attrs['type'])
            py_subcontainer.container_type = py_obj_type

        # Check if we have an unloaded loader for the provided py_obj
        load_loader(py_obj_type)

        if py_subcontainer.container_base_type not in types_not_to_sort:
            h_keys = sort_keys(h_group.keys())
        else:
            h_keys = h_group.keys()

        for h_name in h_keys:
            h_node = h_group[h_name]
            py_subcontainer = _load(py_subcontainer, h_node)

        sub_data = py_subcontainer.convert()
        py_container.append(sub_data)

    else:
        # must be a dataset
        subdata = load_dataset(h_group)
        py_container.append(subdata)

    return py_container



# %% IMPORTS
# Built-in imports
import re

# Package imports
import dill as pickle


# %% FUNCTION DEFINITIONS
def get_type(h_node):
    """ Helper function to return the py_type for an HDF node """
    base_type = h_node.attrs['base_type']
    if base_type != b'pickle':
        py_type = pickle.loads(h_node.attrs['type'])
    else:
        py_type = None
    return py_type, base_type


def get_type_and_data(h_node):
    """ Helper function to return the py_type and data block for an HDF node"""
    py_type, base_type = get_type(h_node)
    data = h_node[()]
    return py_type, base_type, data


def sort_keys(key_list):
    """ Take a list of strings and sort it by integer value within string

    Args:
        key_list (list): List of keys

    Returns:
        key_list_sorted (list): List of keys, sorted by integer
    """

    # Py3 h5py returns an irritating KeysView object
    # Py3 also complains about bytes and strings, convert all keys to bytes
    key_list2 = []
    for key in key_list:
        if isinstance(key, str):
            key = bytes(key, 'ascii')
        key_list2.append(key)
    key_list = key_list2

    # Check which keys contain a number
    numbered_keys = [re.search(br'\d+', key) for key in key_list]

    # Sort the keys on number if they have it, or normally if not
    if(len(key_list) and not numbered_keys.count(None)):
        return(sorted(key_list,
                      key=lambda x: int(re.search(br'\d+', x).group(0))))
    else:
        return(sorted(key_list))


def check_is_iterable(py_obj):
    """ Check whether a python object is a built-in iterable.

    Note: this treats unicode and string as NON ITERABLE

    Args:
        py_obj: python object to test

    Returns:
        iter_ok (bool): True if item is iterable, False is item is not
    """

    # Check if py_obj is an accepted iterable and return
    return(isinstance(py_obj, (tuple, list, set)))


def check_is_hashable(py_obj):
    """ Check if a python object is hashable

    Note: this function is currently not used, but is useful for future
          development.

    Args:
        py_obj: python object to test
    """

    try:
        py_obj.__hash__()
        return True
    except TypeError:
        return False


def check_iterable_item_type(iter_obj):
    """ Check if all items within an iterable are the same type.

    Args:
        iter_obj: iterable object

    Returns:
        iter_type: type of item contained within the iterable. If
            the iterable has many types, a boolean False is returned instead.

    References:
    http://stackoverflow.com/questions/13252333
    """

    iseq = iter(iter_obj)

    try:
        first_type = type(next(iseq))
    except StopIteration:
        return False
    except Exception:   # pragma: no cover
        return False
    else:
        if all([type(x) is first_type for x in iseq]):
            return(first_type)
        else:
            return(False)
